# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

from ... import opcodes as OperandDef
from ...utils import classproperty, TreeReductionBuilder
from .core import DataFrameBinopUfunc, DataFrameArithmeticTreeMixin


class DataFrameOr(DataFrameBinopUfunc):
    _op_type_ = OperandDef.OR

    _bit_func_name = "__or__"
    _bit_rfunc_name = "__ror__"

    @classproperty
    def _operator(self):
        return operator.or_

    @classproperty
    def tensor_op_type(self):
        from ...tensor.arithmetic import TensorBitor

        return TensorBitor


class DataFrameTreeOr(DataFrameArithmeticTreeMixin, DataFrameOr):
    _op_type_ = OperandDef.TREE_OR


def bitor(df, other, axis="columns", level=None, fill_value=None):
    op = DataFrameOr(axis=axis, level=level, fill_value=fill_value, lhs=df, rhs=other)
    return op(df, other)


def rbitor(df, other, axis="columns", level=None, fill_value=None):
    op = DataFrameOr(axis=axis, level=level, fill_value=fill_value, lhs=other, rhs=df)
    return op.rcall(df, other)


def tree_dataframe_or(
    *args, index=None, combine_size=None, axis="columns", level=None, fill_value=None
):
    class MultiplyBuilder(TreeReductionBuilder):
        def _build_reduction(self, inputs, final=False):
            op = DataFrameTreeOr(
                axis=axis,
                level=level,
                fill_value=fill_value,
                output_types=inputs[0].op.output_types,
            )
            params = inputs[0].params.copy()
            params["index"] = index
            return op.new_chunk(inputs, **params)

    return MultiplyBuilder(combine_size).build(args)
