# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from ...serialization.serializables import (
    Serializable,
    FieldTypes,
    StringField,
    ListField,
    BoolField,
    Int32Field,
)
from ...typing import BandType


class SubtaskScheduleSummary(Serializable):
    task_id: str = StringField("task_id")
    subtask_id: str = StringField("subtask_id")
    bands: List[BandType] = ListField("bands", FieldTypes.tuple(FieldTypes.string))
    is_finished: bool = BoolField("is_finished", default=False)
    is_cancelled: bool = BoolField("is_cancelled", default=False)
    num_reschedules: int = Int32Field("num_reschedules", default=0)
