# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

from .... import oscar as mo
from ....lib.aio import alru_cache


class OscarWebAPI:
    def __init__(self, address: str, web_ref: mo.ActorRef):
        self._address = address
        self._web_ref = web_ref

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(cls: Type["OscarWebAPI"], address: str) -> "OscarWebAPI":
        from ..supervisor import WebActor

        ref = await mo.actor_ref(WebActor.default_uid(), address=address)
        return cls(address, ref)

    async def get_web_address(self) -> str:
        return await self._web_ref.get_web_address()
